import getpass
import logging
import os
import re
from configparser import ConfigParser
from urllib.parse import quote

import keyring
import snowflake.connector
from sqlalchemy import create_engine
from .version import __version__
from .sql_utils import to_sql


logger = logging.getLogger(__name__)


def load_env_var(env_var_name, description, print_value=False):
    var = os.getenv(env_var_name)
    if var is None:
        logger.debug(f"Could not find {description} in env var {env_var_name}")
    else:
        if print_value:
            logger.debug(f"Found {description} {var} in env var {env_var_name}")
        else:
            logger.debug(f"Found {description} in env var {env_var_name}")
    return var


def credd_from_env_vars():
    username = load_env_var("NUVOLOS_USERNAME", "username", print_value=True)
    password = load_env_var("NUVOLOS_SF_TOKEN", "Snowflake token", print_value=False)
    if username is None or password is None:
        return None
    else:
        return {"username": username, "snowflake_access_token": password}


def credd_from_odbc_ini():
    credential_filename = os.getenv(
        "NUVOLOS_CREDENTIAL_FILENAME", os.path.expanduser("~") + "/.odbc.ini"
    )
    credential_section = os.getenv("NUVOLOS_CREDENTIAL_SECTION", "nuvolos")
    # Create engine with credentials
    cred = ConfigParser(interpolation=None)
    if not os.path.exists(credential_filename):
        logger.debug(f"Credentials file {credential_filename} not found")
        return None
    cred.read(credential_filename)
    if not cred.has_section(credential_section):
        logger.debug(
            f"Could not find section '{credential_section}' in odbc.ini file {credential_filename}. "
            f"Please add your Nuvolos Snowflake credentials there "
            f"(your username as 'uid' and your Snowflake access token as 'pwd')."
        )
    try:
        odbc_ini = dict(cred.items(credential_section))
    except:
        odbc_ini = {}

    if "uid" not in odbc_ini:
        logger.debug(
            f"Could not find option 'uid' in the '{credential_section}' "
            f"section of odbc.ini file {credential_filename}. "
            f"Please set it to your Nuvolos username"
        )
        return None
    if "pwd" not in odbc_ini:
        logger.debug(
            f"Could not find option 'pwd' in the '{credential_section}' "
            f"section of odbc.ini file {credential_filename}. "
            f"Please set it to your Nuvolos Snowflake access token"
        )
        return None
    logger.debug(f"Found username and snowflake_access_token in {credential_filename}")
    return {"username": odbc_ini["uid"], "snowflake_access_token": odbc_ini["pwd"]}


def credd_from_secrets():
    username_filename = os.getenv("NUVOLOS_USERNAME_FILENAME", "/secrets/username")
    snowflake_access_token_filename = os.getenv(
        "NUVOLOS_SNOWFLAKE_ACCESSS_TOKEN_FILENAME",
        "/secrets/snowflake_access_token",
    )
    if not os.path.exists(username_filename):
        logger.debug(f"Could not find secret file {username_filename}")
        return None
    if not os.path.exists(snowflake_access_token_filename):
        logger.debug(f"Could not find secret file {snowflake_access_token_filename}")
        return None
    with open(username_filename) as username, open(
        snowflake_access_token_filename
    ) as access_token:
        username = username.readline()
        password = access_token.readline()
        logger.debug(f"Found username and Snowflake access token in /secrets files")
        return {"username": username, "snowflake_access_token": password}


def input_nuvolos_credential():
    # store username & password
    username = getpass.getpass("Please input your Nuvolos username:")
    keyring.set_password("nuvolos", "username", username)

    password = getpass.getpass("Please input your Nuvolos password:")
    keyring.set_password("nuvolos", username, password)


def credd_from_local():
    # retrieve username & password
    username = keyring.get_password("nuvolos", "username")
    password = keyring.get_password("nuvolos", username)
    return {"username": username, "snowflake_access_token": password}


def dbpath_from_file(path_filename):
    if not os.path.exists(path_filename):
        logger.debug(f"Could not find dbpath file {path_filename}")
        return None
    with open(path_filename, "r") as path_file:
        lines = path_file.readlines()
        if len(lines) == 0:
            logger.debug(f"Could not parse dbpath file: {path_filename} is empty.")
            return None
        first_line = lines[0].rstrip()
        if "Tables are not enabled" in first_line:
            raise Exception(
                f"Tables are not enabled for this space, please enable them first"
            )
        # Split at "." character
        # This should have resulted in two substrings
        split_arr = re.split('"."', first_line)
        if len(split_arr) != 2:
            logger.debug(
                f'Could not parse dbpath file: pattern "." not found in {path_filename}. '
                f"Are the names escaped with double quotes?"
            )
            return None
        # Remove the remaining double quotes, as we'll escape those
        db_name = split_arr[0].replace('"', "")
        schema_name = split_arr[1].replace('"', "")
        logger.debug(
            f"Found database = {db_name}, schema = {schema_name} in dbpath file {path_filename}."
        )
        return {"db_name": db_name, "schema_name": schema_name}


def dbpath_from_env_vars():
    db_name = load_env_var("NUVOLOS_DB", "Snowflake database", print_value=True)
    schema_name = load_env_var("NUVOLOS_SCHEMA", "Snowflake schema", print_value=True)
    if db_name is None or schema_name is None:
        return None
    return {"db_name": db_name, "schema_name": schema_name}


def _get_connection_params(username=None, password=None, dbname=None, schemaname=None):
    if username is None and password is None:
        credd = (
            credd_from_secrets()
            or credd_from_env_vars()
            or credd_from_odbc_ini()
            or credd_from_local()
        )
        if credd is None:
            input_nuvolos_credential()
            credd = credd_from_local()

        username = credd["username"]
        password = credd["snowflake_access_token"]
    elif username is not None and password is None:
        raise ValueError(
            "You have provided a username but not a password. "
            "Please either provide both arguments or leave both arguments empty."
        )
    elif username is None and password is not None:
        raise ValueError(
            "You have provided a password but not a username. "
            "Please either provide both arguments or leave both arguments empty."
        )
    else:
        logger.debug(f"Found username and Snowflake access token as input arguments")

    if dbname is None and schemaname is None:
        path_filename = os.getenv("NUVOLOS_DBPATH_FILE", "/lifecycle/.dbpath")
        dbd = (
            dbpath_from_file(path_filename)
            or dbpath_from_file(".dbpath")
            or dbpath_from_env_vars()
        )
        if dbd is None:
            raise ValueError(
                "Could not find Snowflake database and schema in .dbpath files or env vars. "
                "If you're not using this function from Nuvolos, "
                "please specify the Snowflake database and schema names as input arguments"
            )
        else:
            db_name = dbd["db_name"]
            schema_name = dbd["schema_name"]
    elif dbname is not None and schemaname is None:
        raise ValueError(
            "You have provided a dbname argument but not a schemaname argument. "
            "Please either provide both or provide none of them."
        )
    elif dbname is None and schemaname is not None:
        raise ValueError(
            "You have provided a schemaname argument but not a dbname argument. "
            "Please either provide both or provide none of them."
        )
    else:
        db_name = dbname
        schema_name = schemaname
        logger.debug(f"Found database and schema as input arguments")

    default_snowflake_host = (
        "acstg.eu-central-1" if "STAGING/" in db_name else "alphacruncher.eu-central-1"
    )
    snowflake_host = os.getenv("NUVOLOS_SNOWFLAKE_HOST", default_snowflake_host)
    return username, password, snowflake_host, db_name, schema_name


def get_url(username=None, password=None, dbname=None, schemaname=None):
    """
    Returns an SQLAlchemy connection URL which can be used to create a connection to Nuvolos.
    :param username: Nuvolos user name.
    :param password: Nuvolos password.
    :param dbname: Nuvolos database name from the Connection Guide.
    :param schemaname: Nuvolos schema name from the Connection Guide.
    :return: An SQLAlchemy connection URL representing the Nuvolos connection.
    """
    username, password, snowflake_host, db_name, schema_name = _get_connection_params(
        username=username, password=password, dbname=dbname, schemaname=schemaname
    )
    url = (
        "snowflake://" + quote(username) + ":" + quote(password) + "@" + snowflake_host
    )
    masked_url = (
        "snowflake://" + quote(username) + ":" + "********" + "@" + snowflake_host
    )

    params = (
        "/?database=%22"
        + quote(db_name)
        + "%22"
        + "&schema=%22"
        + quote(schema_name)
        + "%22"
        + "&CLIENT_METADATA_REQUEST_USE_CONNECTION_CTX=TRUE"
        + "&VALIDATEDEFAULTPARAMETERS=TRUE"
    )
    url = url + params
    masked_url = masked_url + params
    logger.debug("Built SQLAlchemy URL: " + masked_url)
    return url


def get_engine(username=None, password=None, dbname=None, schemaname=None):
    """
    Returns an SQLAlchemy Engine object which can be used with Pandas read_sql/to_sql functions.
    :param username: Nuvolos user name.
    :param password: Nuvolos password.
    :param dbname: Nuvolos database name from the Connection Guide.
    :param schemaname: Nuvolos schema name from the Connection Guide.
    :return: A SQLAlchemy Engine object.
    """
    return create_engine(
        url=get_url(username, password, dbname, schemaname),
        echo=False,
        connect_args={
            "QUERY_TAG": f"nuvolos {__version__}",
        },
    )


def get_connection(username=None, password=None, dbname=None, schemaname=None):
    loc_eng = get_engine(username, password, dbname, schemaname)
    return loc_eng.connect()


def get_raw_connection(username=None, password=None, dbname=None, schemaname=None):
    """
    Returns a raw Snowflake Python Connector API Connection object.
    :param username: Nuvolos user name.
    :param password: Nuvolos password.
    :param dbname: Nuvolos database name from the Connection Guide.
    :param schemaname: Nuvolos schema name from the Connection Guide.
    :return: A snowflake.connector.Connection object
    """
    username, password, snowflake_host, db_name, schema_name = _get_connection_params(
        username=username, password=password, dbname=dbname, schemaname=schemaname
    )
    return snowflake.connector.connect(
        user=username,
        password=password,
        account=snowflake_host,
        database=db_name,
        schema=schema_name,
        session_parameters={
            "QUERY_TAG": f"nuvolos {__version__}",
        },
    )



from setuptools import setup, find_packages


def readme():
    with open("README.rst") as f:
        return f.read()


exec(open('nuvolos/version.py').read())
setup(
    name="nuvolos",
    version=__version__,
    description="The Nuvolos python library for database connectivity",
    long_description=readme(),
    url="https://github.com/nuvolos-cloud/python-connector",
    author="Alphacruncher",
    author_email="support@nuvolos.cloud",
    license="MIT",
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=["keyring", "sqlalchemy>='1.4.0'", "snowflake-connector-python[pandas]<2.6.0", "snowflake-sqlalchemy"],
    zip_safe=False,
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
)



"""
Utility functions to load Pandas DataFrames to Nuvolos/Snowflake.
"""
import os
import random
import string
import re
import logging
from tempfile import TemporaryDirectory
import pandas._libs.lib as lib
from snowflake.connector.errors import ProgrammingError
from sqlalchemy.engine import Connection

from typing import TypeVar, Iterator, Tuple, Sequence

from pandas.core.api import (
    DataFrame,
    Series,
)


logger = logging.getLogger(__name__)
RESERVED_WORDS = frozenset(
    [
        "ALL",  # ANSI Reserved words
        "ALTER",
        "AND",
        "ANY",
        "AS",
        "BETWEEN",
        "BY",
        "CHECK",
        "COLUMN",
        "CONNECT",
        "COPY",
        "CREATE",
        "CURRENT",
        "DELETE",
        "DISTINCT",
        "DROP",
        "ELSE",
        "EXISTS",
        "FOR",
        "FROM",
        "GRANT",
        "GROUP",
        "HAVING",
        "IN",
        "INSERT",
        "INTERSECT",
        "INTO",
        "IS",
        "LIKE",
        "NOT",
        "NULL",
        "OF",
        "ON",
        "OR",
        "ORDER",
        "REVOKE",
        "ROW",
        "ROWS",
        "SAMPLE",
        "SELECT",
        "SET",
        "START",
        "TABLE",
        "THEN",
        "TO",
        "TRIGGER",
        "UNION",
        "UNIQUE",
        "UPDATE",
        "VALUES",
        "WHENEVER",
        "WHERE",
        "WITH",
        "REGEXP",
        "RLIKE",
        "SOME",  # Snowflake Reserved words
        "MINUS",
        "INCREMENT",  # Oracle reserved words
    ]
)
UNQUOTED_RE = re.compile(r"""^[a-z_]+[a-z0-9_\$]*$""")


def _quote_name(name):
    """
    Quotes a Snowflake name if required:
     - The name starts with a non-alpha character,
     - The name contains a special character,
     - The name is a reserved name.
    :param name: Name to quote.
    :return: The quoted name, if quoting is required.
    """
    if name.upper() in RESERVED_WORDS:
        return f'"{name}"'
    elif not UNQUOTED_RE.match(name):
        name = name.replace('"', "")
        return f'"{name}"'
    else:
        return name


def _get_col_db_type(col):
    """
    Returns the Snowflake type for the DataFrame column type.
    See https://pandas.pydata.org/pandas-docs/stable/user_guide/basics.html#basics-dtypes
    Based on https://github.com/pandas-dev/pandas/blob/d01561fb7b9ad337611fa38a3cfb7e8c2faec608/pandas/io/sql.py#L1142:9
    :param col: The Pandas DataFrame column.
    :return: The Snowflake SQL type matching the dtype
    """
    col_type = lib.infer_dtype(col, skipna=True)

    if col_type == "datetime64" or col_type == "datetime":
        # GH 9086: TIMESTAMP is the suggested type if the column contains
        # timezone information
        try:
            if col.dt.tz is not None:
                return "TIMESTAMPT_TZ"
        except AttributeError:
            # The column is actually a DatetimeIndex
            # GH 26761 or an Index with date-like data e.g. 9999-01-01
            if getattr(col, "tz", None) is not None:
                return "TIMESTAMPT_TZ"
        return "TIMESTAMP_NTZ"
    if col_type == "timedelta64":
        logger.warning(
            "the 'timedelta' type is not supported, and will be "
            "written as integer values (ns frequency) to the database.",
        )
        return "INT"
    elif col_type == "floating":
        return "FLOAT"
    elif col_type == "integer":
        return "INT"
    elif col_type == "boolean":
        return "BOOLEAN"
    elif col_type == "date":
        return "DATE"
    elif col_type == "time":
        return "TIME"
    elif col_type == "complex":
        raise ValueError("Complex datatypes not supported")
    return "TEXT"


def _index_name(frame, index, index_label):
    # for writing: index=True to include index in sql table
    if index is True:
        nlevels = frame.index.nlevels
        # if index_label is specified, set this as index name(s)
        if index_label is not None:
            if not isinstance(index_label, list):
                index_label = [index_label]
            if len(index_label) != nlevels:
                raise ValueError(
                    "Length of 'index_label' should match number of "
                    f"levels, which is {nlevels}"
                )
            else:
                return index_label
        # return the used column labels for the index columns
        if nlevels == 1 and "index" not in frame.columns and frame.index.name is None:
            return ["index"]
        else:
            return [
                l if l is not None else f"level_{i}"
                for i, l in enumerate(frame.index.names)
            ]

    # for reading: index=(list of) string to specify column to set as index
    elif isinstance(index, str):
        return [index]
    elif isinstance(index, list):
        return index
    else:
        return None


def _get_column_names_and_types(index, frame, dtype_mapper):
    column_names_and_types = []
    if index is not None:
        for i, idx_label in enumerate(index):
            idx_type = dtype_mapper(frame.index._get_level_values(i))
            column_names_and_types.append((str(idx_label), idx_type, True))

    column_names_and_types += [
        (str(frame.columns[i]), dtype_mapper(frame.iloc[:, i]), False)
        for i in range(len(frame.columns))
    ]
    logger.debug(f"Snowflake types assigned to pandas dtypes: {column_names_and_types}")
    return column_names_and_types


def _get_create_table_statement(table, index, frame, database=None, schema=None):
    """
    Returns a CREATE TABLE statement for the table to which the DataFrame will be loaded.
    :param table: The name of the table which will be created.
    :param index: DataFrame indexes which will be created as DataFrame columns
    :param frame: The DataFrame to load.
    :param database: The name of the database to which data will be inserted.
    :param schema: The name of the schema to which data will be inserted.
    :return: The CREATE TABLE statement.
    """
    column_names_and_types = _get_column_names_and_types(
        index=index, frame=frame, dtype_mapper=_get_col_db_type
    )
    col_defs = []
    for col_name, col_type, is_index in column_names_and_types:
        col_defs.append(f"  {_quote_name(col_name)} {col_type}")
    qualified_name = _qualify_name(database, schema, table)
    return (
        f"CREATE OR REPLACE TABLE {qualified_name} (\n" + ",\n".join(col_defs) + "\n);"
    )


def _qualify_name(database, schema, table):
    if database is not None and schema is None:
        raise ValueError(
            f"Schema is not specified for database [{database}] to create table [{_quote_name(table)}] in."
        )
    qualified_name = _quote_name(table)
    if schema is not None:
        qualified_name = f"{_quote_name(schema)}.{qualified_name}"
    if database is not None:
        qualified_name = f"{_quote_name(database)}.{qualified_name}"
    return qualified_name


def _ensure_table_exists(
    con, table, index, frame, if_exists="fail", database=None, schema=None
):
    """
    Checks if the table exists in the current/given schema and creates it based on the DataFrame/indices if necessary.
    :param con: The pre-opened database Connection to use.
    :param table: The name of the database table. It will be quoted and case sensitive if it contains keywords, lower case or special chars.
    :param index: DataFrame indexes which will be created as DataFrame columns
    :param frame: The DataFrame to load.
    :param if_exists: How to behave if the table already exists. {‘fail’, ‘replace’, ‘append’}, default ‘fail’
             * fail: Raise a ValueError.
             * replace: Drop the table before inserting new values.
             * append: Insert new values to the existing table.
    :param database: The name of the database to which data will be inserted.
    :param schema: The name of the schema to which data will be inserted.
    :return: None
    """
    if database is not None and schema is None:
        raise ValueError(
            f"Schema is not specified for database [{database}] to check if table [{_quote_name(table)}] exists."
        )
    qualified_name = ""
    if schema is not None:
        qualified_name = f" IN SCHEMA {_quote_name(schema)}"
    if database is not None:
        qualified_name = f" IN SCHEMA {_quote_name(database)}.{_quote_name(schema)}"

    name_to_search = _quote_name(table).replace('"', "")
    if _quote_name(table) == table:
        name_to_search = table.upper()

    table_exists = False
    logger.info(f"Checking if table [{table}] exists in [{database}].[{schema}]")
    result = con.execute(f"SHOW TERSE TABLES{qualified_name};")
    for row in result:
        if name_to_search == row[1]:
            logger.info(f"Table [{table}] already exists in [{database}].[{schema}]")
            table_exists = True
    if not table_exists or "replace" == if_exists:
        logger.info(
            f"Table [{table}] doesn't exist yet in [{database}].[{schema}], creating..."
        )
        create_statement = _get_create_table_statement(
            table=table, index=index, frame=frame, database=database, schema=schema
        )
        logger.debug(f"Table will be created as:\n{create_statement}")
        con.execute(create_statement)
        logger.info(f"Table [{table}] created successfully.")
    elif if_exists == "fail":
        raise ValueError(f"Table [{table}] already exists in [{database}].[{schema}]")


T = TypeVar("T", bound=Sequence)


def _chunk_helper(lst: T, n: int) -> Iterator[Tuple[int, T]]:
    """Helper generator to chunk a sequence efficiently with current index like if enumerate was called on sequence."""
    for i in range(0, len(lst), n):
        yield int(i / n), lst[i : i + n]


def to_sql(
    df,
    name,
    con,
    database=None,
    schema=None,
    if_exists="fail",
    index=True,
    index_label=None,
    nanoseconds=False,
):
    """
    Load a DataFrame to the specified table in the database.
    Creates the table if it doesn't yet exist, with TEXT/FLOAT/DATE/TIMESTAMP columns as required.
    The name will be case sensitive (quoted) if it contains lowercase or special characters or is a reserved keyword.
    Based on the write_pandas function of snowflake-connector-python:
    https://docs.snowflake.com/en/user-guide/python-connector-api.html#write_pandas
    :param df: The Pandas DataFrame to insert/stage as a table.
    :param name: The name of the database table. It will only be quoted and case sensitive if it contains keywords or special chars.
    :param con: The pre-opened database Connection to use.
    :param database: The name of the database to which data will be inserted.
    :param schema: The name of the schema to which data will be inserted.
    :param if_exists: How to behave if the table already exists. {‘fail’, ‘replace’, ‘append’}, default ‘fail’
             * fail: Raise a ValueError.
             * replace: Drop the table before inserting new values.
             * append: Insert new values to the existing table.
    :param index: bool, default True: Write DataFrame index as a column. Uses index_label as the column name in the table.
    :param index_label: Column label for index column(s). If None is given (default) and index is True, then the index names are used. A sequence should be given if the DataFrame uses MultiIndex.
    :param nanoseconds: If True, nanosecond timestamps will be used to upload the data. Limits timestamp range from 1677-09-21 00:12:43.145224192 to 2262-04-11 23:47:16.854775807.
    :return: Returns the COPY INTO command's results to verify ingestion in the form of a tuple of whether all chunks were
        ingested correctly, # of chunks, # of ingested rows, and ingest's output.
    """
    if not isinstance(con, Connection):
        raise ValueError(f"Provided con object is not an sqlalchemy.engine.Connection instance.")
    if if_exists not in ("fail", "replace", "append"):
        raise ValueError(f"'{if_exists}' is not valid for if_exists")

    if isinstance(df, Series):
        df = df.to_frame()
    elif not isinstance(df, DataFrame):
        raise NotImplementedError(
            "'df' argument should be either a Series or a DataFrame"
        )

    indices = _index_name(df, index, index_label)
    with con.begin():
        _ensure_table_exists(
            con=con,
            table=name,
            index=indices,
            frame=df,
            if_exists=if_exists,
            database=database,
            schema=schema,
        )

        # Create a temporary stage where the DataFrame will be uploaded as a Parquet file.
        stage_name = None  # Forward declaration
        while True:
            try:
                stage_name = "".join(
                    random.choice(string.ascii_lowercase) for _ in range(5)
                )
                create_stage_sql = (
                    "CREATE TEMPORARY STAGE /* Python:nuvolos.to_sql() */ " '"{stage_name}"'
                ).format(stage_name=stage_name)
                logger.debug("Creating temporary stage with '{}'".format(create_stage_sql))
                con.execute(create_stage_sql)
                break
            except ProgrammingError as pe:
                if pe.msg.endswith("already exists."):
                    logger.debug(
                        f"Temporary stage {stage_name} already exists, choosing a different name."
                    )
                    continue
                raise

        with TemporaryDirectory() as tmp_folder:
            for i, chunk in _chunk_helper(df, 10000000):
                chunk_path = os.path.join(tmp_folder, "file{}.txt".format(i))
                # Dump chunk into parquet file
                if nanoseconds:
                    chunk.to_parquet(
                        chunk_path, compression="snappy", index=index, version="2.0"
                    )
                else:
                    chunk.to_parquet(chunk_path, compression="snappy", index=index)
                # Upload parquet file
                upload_sql = (
                    "PUT /* Python:nuvolos.to_sql() */ "
                    "'file://{path}' @\"{stage_name}\" PARALLEL={parallel}"
                ).format(
                    path=chunk_path.replace("\\", "\\\\").replace("'", "\\'"),
                    stage_name=stage_name,
                    parallel=4,
                )
                logger.debug("Uploading Parquet files with '{}'".format(upload_sql))
                con.execute(upload_sql)
                # Remove chunk file
                os.remove(chunk_path)

        db_columns = []
        parquet_columns = []
        column_names_and_types = _get_column_names_and_types(
            index=indices, frame=df, dtype_mapper=_get_col_db_type
        )
        scale = 6
        if nanoseconds:
            scale = 9
        for col_name, col_type, is_index in column_names_and_types:
            if is_index and indices is not None:
                db_columns.append(_quote_name(col_name))
                if col_type.startswith("TIMESTAMP"):
                    parquet_columns.append(
                        f"TO_TIMESTAMP($1:{_quote_name(col_name)}::INT,{scale})"
                    )
                else:
                    parquet_columns.append(f"$1:{_quote_name(col_name)}")
            else:
                db_columns.append(_quote_name(col_name))
                if col_type.startswith("TIMESTAMP"):
                    parquet_columns.append(
                        f"TO_TIMESTAMP($1:{_quote_name(col_name)}::INT,{scale})"
                    )
                else:
                    parquet_columns.append(f"$1:{_quote_name(col_name)}")
        db_columns = ",".join(db_columns)
        parquet_columns = ",".join(parquet_columns)

        # in Snowflake, all parquet data is stored in a single column, $1, so we must select columns explicitly
        # see (https://docs.snowflake.com/en/user-guide/script-data-load-transform-parquet.html)
        copy_into_sql = (
            "COPY INTO {location} /* Python:nuvolos.to_sql() */ "
            "({columns}) "
            'FROM (SELECT {parquet_columns} FROM @"{stage_name}") '
            "FILE_FORMAT=(TYPE=PARQUET COMPRESSION={compression}) "
            "PURGE=TRUE ON_ERROR={on_error}"
        ).format(
            location=_qualify_name(database=database, schema=schema, table=name),
            columns=db_columns,
            parquet_columns=parquet_columns,
            stage_name=stage_name,
            compression="Snappy",
            on_error="ABORT_STATEMENT",
        )
        logger.debug("Copying into table with '{}'".format(copy_into_sql))
        copy_results = con.execute(copy_into_sql).fetchall()
        return (
            all(e[1] == "LOADED" for e in copy_results),
            len(copy_results),
            sum(e[3] for e in copy_results),
            copy_results,
        )
